package com.weltevree.salvo.core.internal;

import java.io.IOException;
import java.net.Socket;
import java.util.Collection;

import com.weltevree.salvo.core.Debug;
import com.weltevree.salvo.core.TimeoutException;
import com.weltevree.salvo.model.IArticle;
import com.weltevree.salvo.model.INewsgroup;
import com.weltevree.salvo.model.IServer;
import com.weltevree.salvo.model.IServerConnection;

public class ServerConnection implements IServerConnection {

	private Exception lastException;

	private Socket socket;

	private String lastResponse;

	private Collection<INewsgroup> newsgroups;

	private boolean possibleResponseAvailable;

	private Server server;

	private String Jbeq;

	private static final int MAXWAIT = 4;

	private static final String TIMEOUTLINE = "591 No input received within "
			+ MAXWAIT + " seconds\r\n";

	private static final String[] TIMEOUTLIST = new String[] {
			"591 No input received within " + MAXWAIT + " seconds \r\n",
			".\r\n" };

	public ServerConnection(Server server) {
		this.server = server;
	}

	public String getResponse() {

		// If no command was send then return the last message
		if (!isPossibleResponseAvailable())
			return getLastResponse();
		else
			setPossibleResponseAvailable(false);

		// Read from the socket
		try {
			StringBuffer buffer = new StringBuffer();
			if (waitForInput()) {
				byte[] response;
				while (!buffer.toString().endsWith("\r\n")) {
					response = new byte[socket.getInputStream().available()];
					socket.getInputStream().read(response);
					buffer.append(new String(response));
				}
			} else
				buffer.append(TIMEOUTLINE);

			Debug.log(this.getClass(), buffer.toString());
			setLastResponse(buffer.toString());
			return new String(buffer.toString());
		} catch (IOException e) {
			setLastException(e);
			return null;
		}
	}

	public String[] getListResponse() {

		// If no command was send then return the last message
		if (!isPossibleResponseAvailable())
			return getLastResponse().split("\r\n");
		else
			setPossibleResponseAvailable(false);

		try {

			StringBuffer buffer = new StringBuffer();

			if (waitForInput()) {
				// FIXME can get an error here instead of a list. Check for that
				// as well a "400 news.xs4all.nl: Idle timeout" was seen here.
				byte[] response;
				while (!buffer.toString().endsWith("\r\n.\r\n")) {

					if (socket.getInputStream().available() == 0) {
						if (!waitForInput())
							break;
					}
					response = new byte[socket.getInputStream().available()];
					// if(response.length == 0)
					// break;

					socket.getInputStream().read(response);
					buffer.append(new String(response));
				}
			} else
				for (String line : TIMEOUTLIST)
					buffer.append(line);

			Debug.log(this.getClass(), buffer.toString());
			setLastResponse(buffer.toString());
			return new String(buffer.toString()).split("\r\n");
		} catch (IOException e) {
			setLastException(e);
			return null;
		}
	}

	/**
	 * Waits for input. Returns true if input found, false if the wait time was
	 * exceeded without input responses.
	 * 
	 * @return true if input is on the queue, false otherwise.
	 */
	private boolean waitForInput() {

		int counter = 0;

		try {
			while (socket.getInputStream().available() == 0) {
				Thread.sleep(100);

				counter++;

				if (counter / 10 > MAXWAIT)
					throw new TimeoutException("Server " + server.getAddress()
							+ " timed out");

				Debug.logn(".");

			}
		} catch (Exception e) {
			setLastException(e);
			return false;
		}

		Debug.log(this.getClass(), "");
		return true;

	}

	public boolean disconnect() {
		sendCommand("quit");
		try {
			Thread.sleep(100);
			socket.close();
		} catch (Exception e) {
		}
		return socket.isClosed();
	}

	public Collection<INewsgroup> getNewsgroups() {

		if (newsgroups == null || newsgroups.size() == 0)
			newsgroups = ServerRegex.ListNewsgroups(this);

		return newsgroups;
	}

	public boolean sendCommand(String command) {

		if (!isConnected()) {
			Debug.log(this.getClass(), "sendCommand needed to reconnect");
			connect(Cuttle.doConvert(Jbeq));
		}

		System.out.println(command);

		try {
			socket.getOutputStream().write((command + "\r\n").getBytes());
		} catch (IOException e) {
			Debug.log(this.getClass(), "Error received, reconnecting");
			connect(Cuttle.doConvert(Jbeq));
			try {
				socket.getOutputStream().write((command + "\r\n").getBytes());
			} catch (IOException e1) {
				Debug.log(this.getClass(), "Could not send command due to "
						+ e1.getMessage());
				setLastException(e);
				return false;
			}
		}

		setPossibleResponseAvailable(true);
		return true;

	}

	public boolean isConnected() {
		if (socket == null)
			return false;
		return socket.isConnected();
	}

	public Exception getLastException() {
		return lastException;
	}

	private void setLastResponse(String lastResponse) {
		this.lastResponse = lastResponse;
	}

	public String getLastResponse() {
		return lastResponse;
	}

	/**
	 * Indicates if responses are to be expected from the server.
	 * 
	 * @return false if no response is expected from the server
	 */
	private boolean isPossibleResponseAvailable() {
		return possibleResponseAvailable;
	}

	/**
	 * Set by commands (e.g. {@link #sendCommand(String)} that trigger responses
	 * from the server. Set by commands that have read the responses from the
	 * server.
	 * 
	 * @param possibleResponseAvailable
	 */
	private void setPossibleResponseAvailable(boolean possibleResponseAvailable) {
		this.possibleResponseAvailable = possibleResponseAvailable;
	}

	public boolean connect(String passWord) {

		if (!getServer().isAnonymous())
			this.Jbeq = Cuttle.doConvert(passWord);

		try {
			socket = new Socket(server.getAddress(), server.getPort());
			socket.setKeepAlive(true);
			setPossibleResponseAvailable(true);
			getResponse();

			//
			// Logon
			if (!getServer().isAnonymous()) {
				sendCommand("authinfo user " + server.getLogin());
				getResponse();

				//
				// Password
				sendCommand("authinfo pass " + passWord);
				getResponse();
			}

			//
			// Mode reader
			sendCommand("mode reader");
			getResponse();

			// Load overview headers
			ServerRegex.loadOverviewHeaders(getServer());

		} catch (Exception e) {
			setLastException(e);
			return false;
		}

		return isConnected();
	}

	private void setLastException(Exception lastException) {
		this.lastException = lastException;
	}

	public IServer getServer() {
		return server;
	}

	public void getBody(IArticle article) {

	}

	public void flush() {
		getResponse();
	}

}
