package com.weltevree.salvo.core.internal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.weltevree.salvo.core.ArticleFactory;
import com.weltevree.salvo.core.Debug;
import com.weltevree.salvo.model.IArticle;
import com.weltevree.salvo.model.INewsgroup;
import com.weltevree.salvo.model.IServer;
import com.weltevree.salvo.model.IServerConnection;
import com.weltevree.salvo.model.SALVO;

/**
 * Contains regular expressions to parse NNTP server output.
 * 
 * @author jongw
 * 
 */
/**
 * @author jongw
 * 
 */
public class ServerRegex {

	/**
	 * Gets a list of newsgroups.
	 * 
	 * @param connection
	 * @return
	 */
	public static Collection<INewsgroup> ListNewsgroups(IServerConnection connection) {

		connection.sendCommand("list newsgroups");
		String[] lines = connection.getListResponse();
		ArrayList<INewsgroup> list = new ArrayList<INewsgroup>(lines.length);
		int ix = 0;
		for (String line : lines) {
			Debug.log(ServerRegex.class, ix++ + "" + line);

			Newsgroup group = null;

			Matcher m = Pattern.compile("([a-z].*?)(\\s(.*)|$)").matcher(line);

			m.find();
			if (m.matches()) {
				group = new Newsgroup(connection.getServer(), m.group(1), m.group(2).trim());

			}

			if (group != null) {
				list.add(group);
				Debug.log(ServerRegex.class, "Group found: " + group.toString());
			}
		}

		return list;

	}

	/**
	 * Gets the newsgroup article id's with the most new article id in element
	 * 0.
	 * 
	 * @param connection
	 * @return
	 */
	public static Collection<IArticle> getArticles(INewsgroup newsgroup, int from, int to) {

		ArrayList<IArticle> result = new ArrayList<IArticle>();

		// Switch to the newsgroup
		IServerConnection connection = newsgroup.getServer().getServerConnection();

		connection.sendCommand("group " + newsgroup.getNewsgroupName());
		connection.flush();

		// Get all the Articles with XOVER
		connection.sendCommand("xover " + from + "-" + to);
		String[] rList = connection.getListResponse();

		// TODO: XOVER Command and/or LIST OVERVIEW.FTM not supported. Will be
		// handled if time so requires.
		if (!rList[0].startsWith("224"))
			return new ArrayList<IArticle>();

		// Load the article contents
		for (int i = 1; i < rList.length; i++) {
			String overviewResult = rList[i];
			if (!overviewResult.startsWith(".")) {
				IArticle article = ArticleFactory.createArticle(overviewResult, newsgroup);
				if (article != null)
					result.add(article);
			}
		}

		// Adjust newsgroup indexes.
		if (result.size() > 0) {

			// If the high watermark is the same as the to parameter the check
			// if the last message was indeed the value in the high watermark.
			// If it is not then adjust the high watermark in the newsgroup.
			if (to == newsgroup.getHighWaterMark()) {
				newsgroup.adjustHighWatermark(result.get(result.size() - 1).getArticleNumber());
			}

			// If the low watermark is the same as the to parameter the check if
			// the first message was indeed the value in the low watermark. If
			// it is not then adjust the low watermark in the newsgroup.
			if (to == newsgroup.getLowWaterMark()) {
				newsgroup.adjustLowWatermark(result.get(0).getArticleNumber());
			}

			// If the high/low watermarks are the same as the from/to parameters
			// then adjust the article count.
			if (to == newsgroup.getHighWaterMark() && from == newsgroup.getLowWaterMark()) {
				newsgroup.adjustArticleCount(result.size());
			}
		}
		return result;
	}

	/**
	 * Loads the overview headers.
	 * 
	 * @param newsgroup
	 * @return false if not successful
	 */
	public static boolean loadOverviewHeaders(IServer server) {
		/*
		 * Get the format of the overview command
		 */
		IServerConnection connection = server.getServerConnection();
		connection.sendCommand("list overview.fmt");
		String[] rList = connection.getListResponse();

		// TODO: XOVER Command and LIST OVERVIEW.FTM not supported. Will be
		// handled if time so requires.
		if (!rList[0].startsWith("215"))
			return false;

		// Load the overview headers
		server.getOverviewHeaders().clear();
		for (String overviewResult : rList) {
			Debug.log(ServerRegex.class, ".Header: " + overviewResult);
			if (!overviewResult.startsWith("215") && !overviewResult.startsWith(".")) {
				server.getOverviewHeaders().add(overviewResult);
			}
		}
		return true;
	}

	/**
	 * This method goes to the server and asks for the active newsgroup
	 * attributes. These attributes are then placed back into the newsgroup.
	 * 
	 * @param server
	 * @param newsgroup
	 */
	public static void setWaterMarks(INewsgroup newsgroup) {

		IServerConnection connection = newsgroup.getServer().getServerConnection();

		connection.sendCommand("group " + newsgroup.getNewsgroupName());
		String response = connection.getResponse();

		// 211 39270 1 39301 eclipse.tools.emf
		// 
		if (response == null || response.startsWith("411")) {
			newsgroup.setAttributes(0, 0, 0);
			return;
		}

		String[] elements = response.split("\\s");
		if (elements.length == 5) {
			newsgroup.setAttributes(Integer.parseInt(elements[1]), Integer.parseInt(elements[2]), Integer
					.parseInt(elements[3]));
			return;
		} else
			newsgroup.setAttributes(0, 0, 0);
	}

	/**
	 * Gets the bodytext from the server.
	 * 
	 * @param connection
	 * @param newsgroup
	 * @param messageId
	 * @return
	 */
	public static String[] getBody(IArticle article) {

		IServerConnection connection = article.getServer().getServerConnection();
		INewsgroup newsgroup = article.getNewsgroup();
		connection.sendCommand("group " + newsgroup.getNewsgroupName());
		connection.getResponse();

		String messageId = article.getMessageId();
		connection.sendCommand("article " + messageId);

		String[] result = connection.getListResponse();
		final StringBuffer buffie = new StringBuffer(" " + SALVO.CRLF);
		int ix = 0;
		for (String line : result) {
			if (ix++ == 0 && line.startsWith("222"))
				Debug.log(ServerRegex.class, "Skipped " + line);
			else
				buffie.append(line + SALVO.CRLF);
		}

		return buffie.toString().split(SALVO.CRLF);

	}

	/**
	 * Gets the bodytext from the server.
	 * 
	 * @param newsgroup
	 * @param articleId
	 * @return
	 */
	public static IArticle fetchArticle(INewsgroup newsgroup, int articleId, int fetchType) {

		IServerConnection connection = newsgroup.getServer().getServerConnection();

		connection.sendCommand("group " + newsgroup.getNewsgroupName());
		connection.getResponse();

		// Get the article with XOVER
		connection.sendCommand("xover " + articleId);
		String[] rList = connection.getListResponse();

		// TODO: XOVER Command and/or LIST OVERVIEW.FTM not supported. Will be
		// handled if time so requires.
		if (!rList[0].startsWith("224"))
			return null;

		// Load the article contents
		if (!rList[1].startsWith(".")) {
			return ArticleFactory.createArticle(rList[1], newsgroup);
		}

		return null;

	}

	/**
	 * Fetch the followups of this article.
	 * 
	 * @param article
	 * @return the follow ups
	 */
	public static Collection<IArticle> getFollowUps(IArticle article) {

		Collection<IArticle> result = new ArrayList<IArticle>();

		// Switch to the newsgroup
		INewsgroup newsgroup = article.getNewsgroup();
		IServerConnection connection = newsgroup.getServer().getServerConnection();

		connection.sendCommand("group " + newsgroup.getNewsgroupName());
		connection.flush();

		// Get all the follow ups with XPAT
		String reference = article.getMessageId();
		connection.sendCommand("xpat references " + newsgroup.getLowWaterMark() + "-"
				+ newsgroup.getHighWaterMark() + " " + reference);

		String[] rList = connection.getListResponse();

		if (!rList[0].startsWith("221"))
			return new ArrayList<IArticle>();

		// Load the article contents
		for (int i = 1; i < rList.length; i++) {
			String xpatResult = rList[i];
			if (!xpatResult.startsWith(".")) {
				StringTokenizer tizer = new StringTokenizer(xpatResult, " ");
				int range = new Integer(tizer.nextToken()).intValue();
				result.addAll(getArticles(newsgroup, range, range));
			}
		}
		return result;
	}

	public static boolean replyToArticle(IArticle article, StringBuffer body) {

		INewsgroup newsgroup = article.getNewsgroup();
		IServer server = newsgroup.getServer();
		IServerConnection connection = server.getServerConnection();
		connection.flush();

		// Post
		connection.sendCommand("post ");
		String response = connection.getResponse();
		if (!response.startsWith("340")) {
			Debug.log(ServerRegex.class, response);
			// FIXME throw something
			return false;
		}

		connection.sendCommand("From: " + server.getEmail() + " (" + server.getUser() + ")");
		connection.sendCommand("Subject: "
				+ (article.getSubject().startsWith("Re: ") ? article.getSubject() : "Re: "
						+ article.getSubject()));
		connection.sendCommand("X-Organization: Salvo");
		connection.sendCommand("X-Newsreader: Salvo");
		connection.sendCommand("Newsgroups: " + newsgroup.getNewsgroupName());
		StringBuffer references = new StringBuffer();
		for (String ref : article.getReferences()) {
			references.append(ref + " ");
		}
		references.append(article.getMessageId());
		connection.sendCommand("References: " + references.toString());
		connection.sendCommand("");
		for (String line : body.toString().split(SALVO.CRLF)) {
			sendIn80Characters(connection, line);
		}
		connection.sendCommand(".");

		// Get the post response
		response = connection.getResponse();
		System.out.println(response);
		Debug.log(ServerRegex.class, response);
		if (!response.startsWith("240")) {
			return false;
		}

		return true;

	}

	private static void sendIn80Characters(IServerConnection connection, String line) {

		StringBuffer buffer = new StringBuffer();
		StringTokenizer tizer = new StringTokenizer(line, " ");
		while (tizer.hasMoreElements()) {
			String word = tizer.nextToken();
			if ((buffer.length() + word.length() + 1) > SALVO.LINE_LENGTH) {
				connection.sendCommand(buffer.toString());
				buffer.setLength(0);
			}
			buffer.append(word + " ");
		}
		connection.sendCommand(buffer.toString());
	}

	public static boolean postNewArticle(INewsgroup[] newsgroups, String subject, StringBuffer body) {

		INewsgroup newsgroup = newsgroups[0];
		IServer server = newsgroup.getServer();
		IServerConnection connection = server.getServerConnection();
		connection.flush();

		// Post
		connection.sendCommand("post ");
		String response = connection.getResponse();
		if (!response.startsWith("340")) {
			Debug.log(ServerRegex.class, response);
			// FIXME throw something
			return false;
		}

		connection.sendCommand("From: " + server.getEmail() + " (" + server.getUser() + ")");
		connection.sendCommand("Subject: " + subject);
		connection.sendCommand("X-Organization: Salvo");
		connection.sendCommand("X-Newsreader: Salvo");
		StringBuffer toGroups = new StringBuffer();
		for (INewsgroup group : newsgroups) {
			toGroups.append(group.getNewsgroupName());
			toGroups.append(",");
		}
		connection.sendCommand("Newsgroups: " + toGroups.substring(0, toGroups.length() - 1));
//		StringBuffer references = new StringBuffer();
//		connection.sendCommand("References: " + references.toString());
		connection.sendCommand("");
		for (String line : body.toString().split(SALVO.CRLF)) {
			sendIn80Characters(connection, line);
		}
		connection.sendCommand(".");

		// Get the post response
		response = connection.getResponse();
		System.out.println(response);
		Debug.log(ServerRegex.class, response);
		if (!response.startsWith("240")) {
			return false;
		}

		return true;
	}
}
