package com.weltevree.salvo.model;

import java.util.Collection;

/**
 * This class provides methods that both the server side and the store side
 * need. The server side needs to fetch the information and the store side needs
 * to store the information.
 * 
 * @author wimj@weltevree.com
 * 
 */
public interface IInputOutputSystem {

	/**
	 * Posts an article
	 * 
	 * @param article
	 * @return
	 */
	boolean postArticle(IArticle article);

	/**
	 * Cancels an article
	 * 
	 * @param article
	 * @return
	 */
	boolean cancelArticle(IArticle article);

	/**
	 * Gets the article from the newsgroup or the store based on the passed
	 * articleId and the fetchType.
	 * 
	 * @param newsgroup
	 * @param articleId
	 *            the group article id which is used in combination with the
	 *            fetchType.
	 * @param fetchType
	 *            which is one of the fetch types: {@link SALVO#FETCH_NEWEST},
	 *            {@link SALVO#FETCH_NEXT_NEWER}, {@link SALVO#FETCH_NEXT_OLDER}
	 *            , {@link SALVO#FETCH_OLDEST}, {@link SALVO#FETCH_THIS}
	 * @return the article or null if it was not found. Get the last
	 */
	public IArticle fetchArticle(INewsgroup newsgroup, int articleId, int fetchType);

	/**
	 * This returns a list with all follow ups for the give article, without
	 * body. It is most likely always fetched from the store. Otherwise, the
	 * server must know a way to retrieve all follow ups of a certain article
	 * which could be done like so
	 * 
	 * <pre>
	 * xpat references &lt;ebc63c0fc0754033034c7d8abe9ba432$1@www.eclipse.org&gt;
	 * </pre>
	 * 
	 * Yours truly was advised not to use this command by one of the writers of
	 * the inn news server because of incompatibility issues across servers. If
	 * always the latest articles are fetched then the Store should be able to
	 * supply the follow ups.
	 * 
	 * @param article
	 * @return the collection of articles
	 */
	public Collection<IArticle> getFollowUps(IArticle article);

	/**
	 * This returns a list with all articles, without body, of the specified
	 * newsgroup in the specified range. It is either fetched from the store or
	 * from the server if the store does not have the articles. If the articles
	 * must be fetched from the server then this is a good moment to store the
	 * articles in the Store. Please keep in mind that this method is dangerous
	 * in terms of memory use. Implementors are allowed to return a portion of the
	 * requested articles. It should be attempted to at least return
	 * {@link SALVO#BATCH_SIZE} articles.
	 * 
	 * @param newsgroup
	 * @param from
	 * @param to
	 * @return the collection of articles.
	 */
	public Collection<IArticle> getArticles(INewsgroup newsgroup, int from, int to);

	/**
	 * Stores this newsgroup in the list of groups this user subscribes to in
	 * this server and adds this group to the passed server instance by calling
	 * {@link IServer#subscribeNewsgroup(INewsgroup)}. If the group was added to
	 * the list of groups an {@link SALVO#EVENT_ADD_GROUP} event is fired.
	 * 
	 * @param server
	 *            may not be null
	 * @param group
	 *            may not be null
	 * @return true if the group could be stored or this group was already
	 *         stored and false if this could not be done. In the latter case,
	 *         if you suspect a server error, see {@link #getLastException()}
	 */
	public boolean subscribeNewsgroup(INewsgroup group);

	/**
	 * Subscribes to this server. The back end is responsible of storing the
	 * password in a correct manner.
	 * 
	 * @param server
	 * @param passWord
	 * @return true if the server could be stored and false if this could not be
	 *         done. In the latter case, see {@link #getLastException()}
	 */
	public boolean subscribeServer(IServer server, String passWord);

	/**
	 * Removes this newsgroup from the list of groups this user subscribes to in
	 * the server. You should also removes this group from the server by calling
	 * {@link IServer#unsubscribeNewsgroup(ISubscribedNewsgroup)}.
	 * 
	 * @param group
	 *            may not be null
	 * @param permanent
	 *            you may clear irreversible
	 * @return true if the group could be unsubscribed and false if this could
	 *         not be done. In the latter case, see {@link #getLastException()}
	 */
	public boolean unsubscribeNewsgroup(INewsgroup group, boolean permanent);

	/**
	 * Removes this server from the list of servers this user subscribes to.
	 * 
	 * @param server
	 *            may not be null
	 * @param permanent
	 *            true to permanently remove the server or false to keep the
	 *            historical data from which this server can be resurrected.
	 * @return true if the group could be stored and false if this could not be
	 *         done. In the latter case, see {@link #getLastException()}
	 */
	public boolean unsubscribeServer(IServer server, boolean permanent);

	/**
	 * Refreshes the attributes from the server and places it in the passed
	 * newsgroup or moves the information from the passed newsgroup into the
	 * store.
	 * 
	 * @return true if the group could be stored and false if this could not be
	 *         done. In the latter case, see {@link #getLastException()}
	 */
	public boolean updateAttributes(INewsgroup newsgroup);

	/**
	 * Updates the article in the store.
	 * 
	 * @return true if the article could be stored and false if this could not
	 *         be done. In the latter case, see {@link #getLastException()}
	 */
	public void updateArticle(IArticle article);

	/**
	 * Retrieves the body of this article.
	 * 
	 * @param article
	 * @return
	 */
	public String[] getArticleBody(IArticle article);

}
