package com.weltevree.salvo.model;

import java.util.Collection;

public interface IStore extends IStoreEventProvider, IInputOutputSystem {

	/**
	 * Retrieves the list of newsgroups this user is subscribed of the passed
	 * {@link IServer}.
	 * 
	 * @param server
	 * @return a Collection of newsgroups, can be empty may not be null.
	 */
	public Collection<INewsgroup> getSubscribedNewsgroups(IServer server);

	/**
	 * Retrieves the article from the back end.
	 * 
	 * @param newsGroup
	 * @param atricleId
	 * @return the article if it exists in this store or null if it cannot be
	 *         found.
	 */
	public IArticle getArticle(IServer server, INewsgroup newsGroup,
			int articleId);

	/**
	 * @return The last exception or null if there was none.
	 */
	public Exception getLastException();

	/**
	 * Fetches the indicated article from the store.
	 * 
	 * @param newsgroup
	 * @param articleId
	 * @param fetchType
	 *            any of the FETCH_ integers from class {@link SALVO}
	 * @return
	 */
	IArticle fetchArticle(INewsgroup newsgroup, int articleId, int fetchType);

	/**
	 * Adds a group to the list of subscribed groups.
	 * 
	 * @param group
	 */
	public boolean subscribeNewsgroup(INewsgroup group);

	/**
	 * Removes a group from the list of subscribed groups. Please note that
	 * removing a group to the list of subscribed groups does not automatically
	 * remove it from the back end. Please see the javadoc of
	 * {@link IStore#unsubscribeNewsgroup(IServer, INewsgroup)} method for more
	 * information.
	 * 
	 * @param group
	 */
	public boolean unsubscribeNewsgroup(INewsgroup group);

	/**
	 * Move the articles to the store.
	 * 
	 * @param articles
	 * @return false in case it could not be done.
	 */
	public boolean storeArticles(Collection<IArticle> articles);

	/**
	 * Stores the article in the store.
	 * 
	 * @param body
	 * @return false in case it could not be done.
	 */
	public boolean storeArticleBody(IArticle article, String[] body);

	/**
	 * Retrieves a list with all the servers this client subscribes to. This is
	 * store only.
	 * 
	 * @return
	 */
	public Collection<IServer> getSubscribedServers();

	/**
	 * Use this method to check if the store is in sync with the server.
	 * 
	 * @return the last (newest) article in the store or null
	 */
	public IArticle getLastArticle(INewsgroup newsgroup);

	/**
	 * Use this method to check if the store is in sync with the server.
	 * 
	 * @return the first (oldest) article in the store or null
	 */
	public IArticle getFirstArticle(INewsgroup newsgroup);

	/**
	 * Returns a meaningful description of this store. <br>
	 * <br>
	 * Examples:
	 * 
	 * <pre>
	 * Local filesystem storage
	 * Derby 4.2.1
	 * </pre>
	 * 
	 * @return String
	 */
	public String getDescription();

}
