package com.weltevree.salvo.store.filesystem.internal;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.StringTokenizer;

import org.eclipse.equinox.security.storage.ISecurePreferences;
import org.eclipse.equinox.security.storage.SecurePreferencesFactory;
import org.eclipse.equinox.security.storage.StorageException;

import com.weltevree.salvo.core.ArticleFactory;
import com.weltevree.salvo.core.Debug;
import com.weltevree.salvo.core.NewsgroupFactory;
import com.weltevree.salvo.core.ServerFactory;
import com.weltevree.salvo.model.IArticle;
import com.weltevree.salvo.model.INewsgroup;
import com.weltevree.salvo.model.IServer;
import com.weltevree.salvo.model.IStore;
import com.weltevree.salvo.model.IStoreEvent;
import com.weltevree.salvo.model.IStoreEventListener;
import com.weltevree.salvo.model.SALVO;
import com.weltevree.salvo.model.StoreEvent;

/**
 * This implementation of IStore uses the Java file system to store the
 * newsgroup data.
 * 
 * @author Wim Jongman
 * 
 */
public class Store implements IStore {

	private static final String ID_DIR = "\\ids";

	private static final String NUMBERS_DIR = "\\numbers";

	/**
	 * Stores the last exception.
	 */
	private Exception lastException;

	private HashMap<Integer, ArrayList<IStoreEventListener>> listeners;

	private HashMap<String, IServer> storedServers;

	private HashMap<String, Collection<INewsgroup>> subscribedGroups = new HashMap<String, Collection<INewsgroup>>();

	private HashMap<String, IArticle> lastArticles = new HashMap<String, IArticle>();

	private HashMap<String, IArticle> firstArticles = new HashMap<String, IArticle>();

	private String root = "";

	public Store(String root) {
		this.root = root + "\\";
	}

	private ArrayList<String> getServersFromDisk() {

		File file = new File(SALVO.SALVO_HOME + "\\servers.txt");
		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException e) {
				lastException = e;
				return null;
			}
		}

		ArrayList<String> result = new ArrayList<String>();

		FileReader reader = null;
		try {
			reader = new FileReader(file);
			StringBuffer lines = new StringBuffer();
			while (reader.ready()) {
				lines.append((char) reader.read());
			}

			StringTokenizer tizer = new StringTokenizer(lines.toString(), "\n");

			while (tizer.hasMoreTokens()) {
				String line = tizer.nextToken();
				if (!line.trim().startsWith("#")) {
					result.add(line);
				}
			}

		} catch (Exception e) {
			lastException = e;
			return null;
		}

		finally {

			if (reader != null)
				try {
					reader.close();
				} catch (IOException e) {
					lastException = e;
					return null;
				}
		}

		return result;
	}

	private void loadSubscribedGroups(IServer server) {

		File file = new File(SALVO.SALVO_HOME + "\\" + server.getAddress()
				+ "\\groups.txt");
		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException e) {
				lastException = e;
			}
		}

		FileReader reader = null;
		try {
			reader = new FileReader(file);
			StringBuffer lines = new StringBuffer();
			while (reader.ready()) {
				lines.append((char) reader.read());
			}

			StringTokenizer tizer = new StringTokenizer(lines.toString(), "\n");

			while (tizer.hasMoreTokens()) {
				INewsgroup group = NewsgroupFactory.createNewsGroup(server,
						tizer.nextToken(), "");
				getSubscribedNewsgroups(server).add(group);
				Debug.log(this.getClass(), "Adding group subscription: "
						+ group.toString());
			}

		} catch (Exception e) {
			Debug.log(this.getClass(), e.getMessage());
			lastException = e;
		}

		finally {

			if (reader != null)
				try {
					reader.close();
				} catch (IOException e) {
					lastException = e;
				}
		}
	}

	/**
	 * Writes the subscribed groups in the groups.txt file in the server home
	 * directory. If the server directory does not exist it is created.
	 * 
	 * @param server
	 */
	private void writeSubscribedGroups(IServer server) {

		// Create the server in this database and create a reference to the
		// newsgroups in groups.txt
		File file = new File(getServerHome(server) + "\\groups.txt");
		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException e) {
				Debug.log(getClass(), e);
				lastException = e;
			}
		}

		// Write the name of the newsgroup into the file, create a directory
		// for the newsgroup and create the group info file.
		try {
			FileWriter writer = new FileWriter(file);
			for (INewsgroup group : getSubscribedNewsgroups(server)) {
				writer.write(group.getNewsgroupName() + "\n");
			}
			writer.close();

		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
		}
	}

	private void writeSubscribedServers() {

		File file = new File(SALVO.SALVO_HOME + "\\servers.txt");
		if (file.exists())
			file.delete();

		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException e) {
				Debug.log(getClass(), e);
				lastException = e;
			}
		}

		try {
			FileWriter writer = new FileWriter(file);
			for (IServer server : storedServers.values()) {
				writer.write(server.toString() + "\n");
			}
			writer.close();

		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
		}

	}

	public boolean subscribeServer(final IServer server, final String passWord) {

		ISecurePreferences prefs = SecurePreferencesFactory.getDefault();
		ISecurePreferences node = prefs.node("/com/weltevree/salvo");

		try {
			node.put(server.getAddress(), passWord, true);
		} catch (StorageException e) {
			Debug.log(getClass(), e);
			lastException = e;
			return false;
		}

		storedServers.put(server.getAddress(), server);

		writeSubscribedServers();

		fireEvent(new StoreEvent(server, SALVO.EVENT_ADD_SERVER));

		return true;
	}

	public Exception getLastException() {
		return lastException;
	}

	public void addListener(IStoreEventListener listener, int eventType) {

		if (listeners == null) {
			listeners = new HashMap<Integer, ArrayList<IStoreEventListener>>();
		}

		synchronized (listeners) {

			ArrayList<IStoreEventListener> list = listeners.get(new Integer(
					eventType));
			if (list == null) {
				list = new ArrayList<IStoreEventListener>();
				listeners.put(new Integer(eventType), list);
			}
			if (!list.contains(listener))
				list.add(listener);
		}
	}

	public void fireEvent(IStoreEvent event) {

		if (listeners == null || listeners.isEmpty())
			return;

		synchronized (listeners) {

			for (Integer key : listeners.keySet()) {
				if ((event.getEventType() | key.intValue()) == key.intValue()) {
					Debug.log(getClass(), "Listeners found for event type "
							+ key.intValue() + " (" + event.getEventType()
							+ "): ");

					ArrayList<IStoreEventListener> list = listeners.get(key);
					for (IStoreEventListener listener : list) {
						Debug.log(getClass(), "Calling Listener "
								+ listener.getClass().getSimpleName());
						listener.storeEvent(event);
					}
				}
			}
		}
	}

	public void removeListener(IStoreEventListener listener) {

		if (listeners == null || listeners.isEmpty())
			return;
		for (Integer key : listeners.keySet()) {
			ArrayList<IStoreEventListener> list = listeners.get(key);
			if (list.contains(listener)) {
				list.remove(listener);
				if (list.isEmpty()) {
					listeners.remove(key);
				}
			}
		}
	}

	/**
	 * Returns the home directory of this server and creates it if it does not
	 * exist.
	 * 
	 * @param server
	 * @return
	 */
	public String getServerHome(IServer server) {

		File file = new File(SALVO.SALVO_HOME + "\\" + root
				+ server.getAddress());
		if (!file.exists())
			file.mkdirs();
		return file.getAbsolutePath();

	}

	/**
	 * Returns the home directory of this newsgroup and creates it if it does
	 * not exist.
	 * 
	 * @param server
	 * @return
	 */
	public String getNewsgroupHome(INewsgroup newsgroup) {

		String home = getServerHome(newsgroup.getServer()) + "\\"
				+ newsgroup.getNewsgroupName();

		File file = new File(home);
		if (!file.exists()) {
			file.mkdirs();
		}

		File numberDir = new File(home + NUMBERS_DIR);
		if (!numberDir.exists()) {
			numberDir.mkdirs();
		}

		File idDir = new File(home + ID_DIR);
		if (!idDir.exists()) {
			idDir.mkdirs();
		}

		return file.getAbsolutePath();

	}

	public boolean unsubscribeNewsgroup(INewsgroup group, boolean permanent) {

		unsubscribeNewsgroup(group);
		writeSubscribedGroups(group.getServer());

		// remove newsgroup directory from disk
		if (permanent) {
			File file = new File(getNewsgroupHome(group));
			if (file.exists())
				file.delete();
		}

		fireEvent(new StoreEvent(group, SALVO.EVENT_REMOVE_GROUP));
		return true;
	}

	public boolean unsubscribeServer(IServer server, boolean permanent) {
		// Disconnect first
		server.getServerConnection().disconnect();

		// Unsubscribe all groups
		for (INewsgroup group : getSubscribedNewsgroups(server)) {
			unsubscribeNewsgroup(group, permanent);
			fireEvent(new StoreEvent(group, SALVO.EVENT_REMOVE_GROUP));
		}

		// remove server directory from disk
		if (permanent) {
			File file = new File(getServerHome(server));
			if (file.exists())
				file.delete();
		}

		// remove this server from the preferences
		ISecurePreferences prefs = SecurePreferencesFactory.getDefault();
		ISecurePreferences node = prefs.node("/com/weltevree/salvo");
		try {
			node.remove(server.getAddress());
		} catch (IllegalStateException e) {
			Debug.log(getClass(), e);
			lastException = e;
			return false;
		}

		// Remove the server from the list
		storedServers.remove(server.getAddress());

		// Remove it from the servers file
		writeSubscribedServers();

		// Notify
		fireEvent(new StoreEvent(server, SALVO.EVENT_REMOVE_SERVER));

		return true;

	}

	public IArticle getArticle(IServer server, INewsgroup newsGroup,
			int articleId) {
		// TODO Auto-generated method stub
		return null;
	}

	public IArticle fetchArticle(INewsgroup newsgroup, int articleId,
			int fetchType) {
		// TODO Auto-generated method stub
		return null;
	}

	public boolean postArticle(IArticle article) {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean subscribeNewsgroup(INewsgroup group) {

		if (getSubscribedNewsgroups(group.getServer()).contains(group)) {
			Debug.log(this.getClass(), "Group " + group + " already added");
			return true;
		}

		getSubscribedNewsgroups(group.getServer()).add(group);
		writeSubscribedGroups(group.getServer());
		fireEvent(new StoreEvent(group, SALVO.EVENT_ADD_GROUP));
		Debug.log(this.getClass(), "Group " + group + " added.");
		return true;
	}

	public boolean unsubscribeNewsgroup(INewsgroup group) {
		if (subscribedGroups.get(group.getServer().getAddress()) != null)
			subscribedGroups.get(group.getServer().getAddress()).remove(group);
		return true;
	}

	public Collection<INewsgroup> getSubscribedNewsgroups(IServer server) {

		if (subscribedGroups.get(server.getAddress()) == null)
			subscribedGroups.put(server.getAddress(),
					new ArrayList<INewsgroup>());
		return subscribedGroups.get(server.getAddress());
	}

	public boolean cancelArticle(IArticle article) {
		// TODO Auto-generated method stub
		return false;
	}

	public boolean updateAttributes(INewsgroup newsgroup) {

		Debug.log(getClass(), "Updating attributes of group " + newsgroup);

		try {

			// Get the old newsgroup from disk
			INewsgroup oldGroup = null;
			File serialized = new File(getNewsgroupHome(newsgroup)
					+ "\\Newsgroup.data");
			oldGroup = NewsgroupFactory.createNewsGroup(serialized);

			// Write back if changed
			if (oldGroup == null
					|| (newsgroup.getLastChangeDate().getTimeInMillis() != oldGroup
							.getLastChangeDate().getTimeInMillis())) {
				FileOutputStream f_out = new FileOutputStream(
						getNewsgroupHome(newsgroup) + "\\Newsgroup.data");
				ObjectOutputStream obj_out = new ObjectOutputStream(f_out);
				obj_out.writeObject(newsgroup);

				// Send store event
				fireEvent(new StoreEvent(newsgroup, SALVO.EVENT_CHANGE_GROUP));
			}
		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
			return false;
		}

		return true;
	}

	public Collection<IServer> getSubscribedServers() {
		if (storedServers != null)
			return storedServers.values();

		storedServers = new HashMap<String, IServer>();

		ArrayList<String> serversOnDisk = getServersFromDisk();

		for (String line : serversOnDisk) {

			StringTokenizer tizer = new StringTokenizer(line, "::");
			String address = tizer.nextToken();
			int port = Integer.valueOf(tizer.nextToken());
			String user = tizer.nextToken();
			String email = tizer.nextToken();
			String logIn = tizer.nextToken();
			boolean secure = Boolean.parseBoolean(tizer.nextToken());

			ISecurePreferences prefs = SecurePreferencesFactory.getDefault();
			ISecurePreferences node = prefs.node("/com/weltevree/salvo");
			String pass;
			try {
				pass = node.get(address, "");
			} catch (Exception e) {
				Debug.log(this.getClass(), "Storage Exception: "
						+ e.getMessage());
				lastException = e;
				return storedServers.values();
			}

			IServer server = ServerFactory.getServer(address, port, user,
					email, logIn, pass, secure);
			loadSubscribedGroups(server);

			storedServers.put(server.getAddress(), server);

			fireEvent(new StoreEvent(server, SALVO.EVENT_RELOAD));

		}

		return storedServers.values();
	}

	public Collection<IArticle> getArticles(INewsgroup newsgroup, int from,
			int to) {

		// If this range is not in store then return null
		IArticle article = getArticle(newsgroup, to);
		if (article == null)
			return null;

		article = getArticle(newsgroup, from);
		if (article == null)
			return null;

		// Read all files
		File numbersDir = new File(getNumbersHome(newsgroup));
		if (!numbersDir.exists()) {
			numbersDir.mkdir();
			return null;
		}

		String fromString = formatNumber(from);
		String toString = formatNumber(to);

		ArrayList<String> files = new ArrayList<String>(to - from);
		for (File file : numbersDir.listFiles()) {
			if (file.getName().compareTo(fromString) >= 0
					&& file.getName().compareTo(toString) <= 0) {
				files.add(file.getName());
			}
		}

		String[] sortedFiles = files.toArray(new String[files.size()]);
		Arrays.sort(sortedFiles, new Comparator<String>() {
			public int compare(String arg0, String arg1) {
				return arg0.compareTo(arg1) * -1;
			}
		});

		// Fill an article array
		int timer = Debug.timerStart(this.getClass());
		ArrayList<IArticle> result = new ArrayList<IArticle>();
		for (String file : sortedFiles) {
			result.add(getArticle(newsgroup, Integer.parseInt(file)));
		}
		Debug.timerStop(getClass(), timer);

		return result;

	}

	/**
	 * Returns the home directory of the numbers dir.
	 * 
	 * @param newsgroup
	 * @return
	 */
	private String getNumbersHome(INewsgroup newsgroup) {
		return getNewsgroupHome(newsgroup) + NUMBERS_DIR;
	}

	/**
	 * Returns the home directory of the id dir.
	 * 
	 * @param newsgroup
	 * @return
	 */
	private String getIdsHome(INewsgroup newsgroup) {
		return getNewsgroupHome(newsgroup) + ID_DIR;
	}

	/**
	 * Gets the article with this number from the newsgroup store.
	 * 
	 * @param newsgroup
	 * @param number
	 * @return the article or null if it does not exist
	 */
	private IArticle getArticle(INewsgroup newsgroup, int number) {

		File file = new File(getArticleFilenameByNumber(newsgroup, number));
		if (!file.exists())
			return null;
		return getArticle(newsgroup, file);
	}

	/**
	 * Returns the name of the file that contains the article information.
	 * 
	 * @param newsgroup
	 * @param number
	 *            the article number
	 * @return
	 */
	private String getArticleFilenameByNumber(INewsgroup newsgroup, int number) {
		return getNumbersHome(newsgroup) + "\\" + formatNumber(number);
	}

	/**
	 * Returns the name of the file that contains the article information.
	 * 
	 * @param newsgroup
	 * @param id
	 *            the article ID
	 * @return
	 */
	private String getArticleFilenameById(INewsgroup newsgroup, String id) {
		return getIdsHome(newsgroup) + "\\" + id.replaceAll("(<|>)", "");
	}

	public boolean storeArticles(Collection<IArticle> articles) {

		boolean result = internalStoreArticles(articles);

		// Send store event
		if (result)
			fireEvent(new StoreEvent(articles, SALVO.EVENT_CHANGE_GROUP));

		return result;
	}

	/**
	 * Does what storeArticles should do. This was internalized so that it can
	 * also be called by update methods.
	 * 
	 * @param articles
	 * @return false if not success else true
	 */
	private boolean internalStoreArticles(Collection<IArticle> articles) {
		FileWriter fileWriter = null;
		File file = null;

		try {

			for (IArticle article : articles) {

				String articleFilename = getArticleFilename(article);
				file = new File(articleFilename);

				// Delete an existing file
				if (file.exists())
					file.delete();

				// Serialize this article object
				FileOutputStream f_out = new FileOutputStream(file);
				ObjectOutputStream obj_out = new ObjectOutputStream(f_out);
				obj_out.writeObject(article);
				obj_out.close();

				// Replace the complete filename with the last part of the
				// filename
				articleFilename = file.getName();

				// Create the number file for this article. It contains the name
				// of the article file.
				file = new File(getArticleFilenameByNumber(article
						.getNewsgroup(), article.getArticleNumber()));
				if (file.exists())
					file.delete();
				fileWriter = new FileWriter(file);
				fileWriter.write(articleFilename);
				fileWriter.close();

				// Create the id file for this article. It contains the name
				// of the article file.
				file = new File(getArticleFilenameById(article.getNewsgroup(),
						article.getMessageId()));
				if (file.exists())
					file.delete();
				fileWriter = new FileWriter(file);
				fileWriter.write(articleFilename);
				fileWriter.close();

				// Create a folder for the body and the follow up articles.
				file = new File(getArticleDirectory(article.getNewsgroup(),
						article.getMessageId()));
				if (!file.exists())
					file.mkdirs();

				// If this article is a follow up, store it in the main article
				String reference = article.getLastReference();
				if (reference != null) {
					// Create a folder for the body and the follow up articles
					// if it does not yet exist.
					file = new File(getArticleDirectory(article.getNewsgroup(),
							reference));
					if (!file.exists())
						file.mkdirs();

					// Serialize this article object
					f_out = new FileOutputStream(getFollowUpFilename(article,
							reference));
					obj_out = new ObjectOutputStream(f_out);
					obj_out.writeObject(article);
					obj_out.close();
				}

				// Set as first or last if needed
				setFirstOrLastArticle(article);

			}
		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
			return false;
		}

		finally {

			try {
				if (fileWriter != null)
					fileWriter.close();
			} catch (IOException e) {
				Debug.log(getClass(), e);
				lastException = e;
			}
		}
		return true;
	}

	/**
	 * Returns the name of the directory where the article content is stored.
	 * 
	 * @param newsgroup
	 * @param messageId
	 * @return
	 */
	private String getArticleDirectory(INewsgroup newsgroup, String messageId) {
		return getNewsgroupHome(newsgroup) + "\\"
				+ messageId.replaceAll("(<|>)", "");
	}

	/**
	 * Returns the name of the followup article file name given the article and
	 * the parent
	 * 
	 * @param article
	 * @param parentId
	 * @return
	 */
	private String getFollowUpFilename(IArticle article, String parentId) {
		return getNewsgroupHome(article.getNewsgroup()) + "\\"
				+ parentId.replaceAll("(<|>)", "") + "\\"
				+ article.getMessageId().replaceAll("(<|>)", "") + ".data";
	}

	/**
	 * Possibly replace the first or/and last articles in the store with the
	 * passed article.
	 * 
	 * @param article
	 */
	private void setFirstOrLastArticle(IArticle article) {

		IArticle oldArticle = firstArticles.get(article.getNewsgroup()
				.getUniqueNewsgroupName());
		if (oldArticle == null
				|| oldArticle.getArticleNumber() > article.getArticleNumber()) {
			firstArticles.put(article.getNewsgroup().getUniqueNewsgroupName(),
					article);
		}

		oldArticle = lastArticles.get(article.getNewsgroup()
				.getUniqueNewsgroupName());
		if (oldArticle == null
				|| oldArticle.getArticleNumber() < article.getArticleNumber()) {
			lastArticles.put(article.getNewsgroup().getUniqueNewsgroupName(),
					article);
		}

	}

	/**
	 * 
	 * Gets the filename used to store the article.
	 * 
	 * <pre>
	 * nnnnnn.A.&lt;I&gt;.data 
	 * nnnnnn = Article number in newsgroup 
	 * A = A new post and B is follow up
	 * </pre>
	 * 
	 * If the article is a follow up then it is written again like this
	 * 
	 * <pre>
	 * R.nnnnnn.oooooo.data 
	 * R=R 
	 * nnnnnn = number of article that is the original post 
	 * oooooo = nummber of article that is the reply
	 * </pre>
	 * 
	 * @param article
	 * @return the filename
	 */
	private String getArticleFilename(IArticle article) {

		String[] refs = article.getReferences();
		StringBuffer fileName = new StringBuffer();
		fileName.append(getNewsgroupHome(article.getNewsgroup()) + "\\");
		fileName.append(formatNumber(article.getArticleNumber()));
		if (refs.length == 0)
			fileName.append(".A.");
		else
			fileName.append(".B.");
		fileName.append(article.getMessageId().replaceFirst("<", "")
				.replaceFirst(">", ""));
		fileName.append(".data");

		return fileName.toString();
	}

	/**
	 * Pads until we get six digit string.
	 * 
	 * @param number
	 * @return
	 */
	private String formatNumber(int number) {

		if (number < 10)
			return "00000" + number;
		if (number < 100)
			return "0000" + number;
		if (number < 1000)
			return "000" + number;
		if (number < 10000)
			return "00" + number;
		if (number < 100000)
			return "0" + number;
		return "" + number;
	}

	public IArticle getFirstArticle(INewsgroup newsgroup) {

		if (firstArticles.get(newsgroup.getUniqueNewsgroupName()) != null) {
			return firstArticles.get(newsgroup.getUniqueNewsgroupName());
		}

		File infoDir = new File(getNumbersHome(newsgroup));
		if (!infoDir.exists()) {
			infoDir.mkdir();
			return null;
		}

		File first = new File("999999");
		for (File file : infoDir.listFiles()) {
			if (file.getName().compareTo(first.getName()) < 0) {
				first = file;
			}
		}

		if (first.getName().equals("999999")) {
			return null;
		}

		IArticle firstArticle = getArticle(newsgroup, first);
		firstArticles.put(newsgroup.getUniqueNewsgroupName(), firstArticle);
		return firstArticle;
	}

	/**
	 * Returns the article found in the infoFile file.
	 * 
	 * @param infoFile
	 * @return
	 */
	private IArticle getArticle(INewsgroup newsgroup, File infoFile) {
		try {
			FileReader reader = new FileReader(infoFile);
			int size = Integer.parseInt(infoFile.length() + "");
			char[] fileNameChar = new char[size];
			reader.read(fileNameChar);
			reader.close();
			Object obj = ArticleFactory.createArticle(new File(
					getNewsgroupHome(newsgroup) + "\\"
							+ String.valueOf(fileNameChar)));
			if (obj instanceof IArticle) {
				IArticle article = (IArticle) obj;
				article.setNewsgroup(newsgroup);
				return article;
			}
		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
		}
		return null;
	}

	public IArticle getLastArticle(INewsgroup newsgroup) {

		if (lastArticles.get(newsgroup.getUniqueNewsgroupName()) != null) {
			return lastArticles.get(newsgroup.getUniqueNewsgroupName());
		}

		File infoDir = new File(getNumbersHome(newsgroup));
		if (!infoDir.exists()) {
			infoDir.mkdir();
			return null;
		}

		File last = new File("000000");
		for (File file : infoDir.listFiles()) {
			if (file.getName().compareTo(last.getName()) > 0) {
				last = file;
			}
		}

		if (last.getName().equals("000000")) {
			return null;
		}

		IArticle lastArticle = getArticle(newsgroup, last);
		lastArticles.put(newsgroup.getUniqueNewsgroupName(), lastArticle);
		return lastArticle;
	}

	public String[] getArticleBody(IArticle article) {

		// Get the body from disk
		try {
			File serialized = new File(getArticleDirectory(article
					.getNewsgroup(), article.getMessageId())
					+ "\\body.data");
			if (serialized.exists()) {
				FileInputStream f_in = new FileInputStream(serialized);
				ObjectInputStream obj_in = new ObjectInputStream(f_in);
				Object obj = obj_in.readObject();
				if (obj instanceof String[]) {
					return (String[]) obj;
				}
			}
		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
		}

		// FIXME should we store article bodies in the store?
		return null;
	}

	public boolean storeArticleBody(IArticle article, String[] body) {

		try {
			File file = new File(getArticleDirectory(article.getNewsgroup(),
					article.getMessageId()));
			if (!file.exists())
				file.mkdirs();
			file = new File(file.getAbsolutePath() + "\\body.data");

			// FIXME Should we assume that bodies are immutable?
			if (file.exists())
				return true;

			// Serialize this article body
			FileOutputStream f_out = new FileOutputStream(file);
			ObjectOutputStream obj_out = new ObjectOutputStream(f_out);
			obj_out.writeObject(body);
			obj_out.close();
		} catch (Exception e) {
			Debug.log(getClass(), e);
			lastException = e;
			return false;
		}

		return true;

	}

	public Collection<IArticle> getFollowUps(IArticle article) {

		Collection<IArticle> result = new ArrayList<IArticle>();

		// Get the follow ups from disk
		File articleDir = new File(getArticleDirectory(article.getNewsgroup(),
				article.getMessageId()));
		if (!articleDir.exists())
			return result;
		for (File serialized : articleDir.listFiles(new FileFilter() {

			public boolean accept(File pathname) {
				if (pathname.getAbsolutePath().endsWith("data"))
					return true;
				return false;
			}
		}))

		{

			if (serialized.exists()) {
				Object obj = ArticleFactory.createArticle(serialized);
				if (obj instanceof IArticle) {
					IArticle newArticle = (IArticle) obj;
					newArticle.setNewsgroup(article.getNewsgroup());
					newArticle.setParent(article);
					result.add(newArticle);
				}
			}
		}

		return result;

	}

	public void updateArticle(IArticle article) {
		ArrayList<IArticle> list = new ArrayList<IArticle>();
		list.add(article);
		boolean result = storeArticles(list);

		// Send store event
		if (result)
			fireEvent(new StoreEvent(article, SALVO.EVENT_CHANGE_GROUP));
	}

	public String getDescription() {
		return "Local Filesystem Storage";
	}
}
